/*
 * File:       iTunesPlugIn.c
 *
 * Abstract:   SpectroGraph, a visual plug-in for iTunes
 *
 * Version:    2.0
 *
 * Author:     Alexander Thomas (http://www.dr-lex.be)
 *
 * License: You may modify this source code and publish the modified plug-in, at the
 *          condition that the modified source code is also made available to the public.
 *          This requirement holds as long as the resulting plug-in is similar to
 *          SpectroGraph, i.e. the basic functionality, displaying a moving spectrogram,
 *          possibly with separate colors for left and right channels, is preserved.
 *          If you make improvements to this plug-in, I would like to hear from you so
 *          I can update the SpectroGraph webpage. Please contact me at:
 *          http://www.dr-lex.be/mailform.html?subject=SpectroGraph
 *          This plug-in and its source code are provided without any implied warranties
 *          of fitness for a particular purpose. Use at your own risk.
 *
 * Change History (most recent first):
 *   08/08/16   Lex     Complete rewrite, using OpenGL for rendering, added scrolling
 *   06/12/26   Lex     fixed overflow bug in speed limiter (released 1.2.1)
 *   06/11/25   Lex     added speed limiter, Universal Binary, released SpectroGraph 1.2
 *   06/06/06   Lex     added anti-banding, released SpectroGraph 1.1
 *   04/07/02   Lex     made SpectroGraph source available to public
 *   01/06/20   Lex     released SpectroGraph 1.0 conversion
 *   01/06/06   KG      moved to project builder on Mac OS X
 *   01/04/17   DTS     first checked in.
 *
 * TODO: adapt for Windows. Some parts are provided, but a lot of work remains.
 *
 */

//########################################
//	includes
//########################################

#include <unistd.h>
#include <math.h>
#include "iTunesVisualAPI.h"

#if TARGET_OS_MAC
#include <AGL/agl.h>
#include <OpenGL/gl.h>
#include <OpenGL/glext.h>
#include <OpenGL/glu.h>
#include <sys/time.h> // this may be tricky to get working under Windows
#endif


#if TARGET_OS_WIN32
#define	MAIN iTunesPluginMain
#define IMPEXP	__declspec(dllexport)
#else
#define IMPEXP
#define	MAIN main
#endif

// Define to enable extra tests and console messages
//#define SG_DEBUG

//########################################
//	typedef's, struct's, enum's, etc.
//########################################

#define kTVisualPluginName              "\014SpectroGraph"
#define	kTVisualPluginCreator			'hook'

#define	kTVisualPluginMajorVersion		2
#define	kTVisualPluginMinorVersion		0
#define	kTVisualPluginReleaseStage		finalStage
//#define	kTVisualPluginReleaseStage		developStage
#define	kTVisualPluginNonFinalRelease	0

// Width = frequency resolution (should be 256)
#define SG_TEXWIDTH  256
// Larger textures make glCopyTexSubImage2D slower. 256*128 seems close to
// the point where improvement becomes negligible.
#define SG_TEXHEIGHT 128
// Number of textures to allocate, must be such that SG_NTEXTURES*SG_TEXHEIGHT
// is as wide as the widest possible display. 4096 pixels seems reasonable...
#define SG_NTEXTURES 32
// Maximum number of lines per update (must be power of 2). 4 seems the maximum,
// above this the graph starts showing obvious bands.
#define SG_MAXCHUNK  4


#define SG_USLEEP		1000
#define SG_FACTOR		1.1
/* SG_MINDELAY must be such that SG_FACTOR*SG_MINDELAY >= SG_MINDELAY+1 */
#define SG_MINDELAY		10 // 100000lps
#define SG_FASTDELAY	1000 // 1000lps
#define SG_NORMDELAY	7500  // 133lps
#define SG_SLOWDELAY	100000 // 10lps
#define SG_MAXDELAY		1000000 // 1lps

// Choosing the right types here is crucial for optimal speed of glCopyTexSubImage2D
#if __BIG_ENDIAN__
#define ARGB_IMAGE_TYPE GL_UNSIGNED_INT_8_8_8_8_REV
#else
#define ARGB_IMAGE_TYPE GL_UNSIGNED_INT_8_8_8_8
#endif


// These IDs correspond to the IDs in Interface Builder
enum {
	kColorSettingID  = 1, 
	kInvertSettingID = 2,
	kBandSettingID   = 3,
	kScrollSettingID = 4,
	kDirSettingID    = 5,
	kSpeedSettingID  = 6,
	kOKSettingID     = 7
};

struct VisualPluginData {
	void *				appCookie;
	ITAppProcPtr		appProc;

	#if TARGET_OS_MAC
	CGrafPtr			destPort;
	#else
	HWND				destPort;
	#endif
	Rect				destRect;
	OptionBits			destOptions;
	UInt32				destBitDepth;

	RenderVisualData	renderData;
	UInt32				renderTimeStampID;
	
	ITTrackInfo			trackInfo;
	ITStreamInfo		streamInfo;

	Boolean				playing;
	Boolean				padding[3];

//	Plugin-specific data
	UInt8				minLevel[kVisualMaxDataChannels];		// 0-128
	UInt8				maxLevel[kVisualMaxDataChannels];		// 0-128

	UInt8				min, max;
};
typedef struct VisualPluginData VisualPluginData;

#if TARGET_OS_MAC
static AGLContext myContext = NULL;
// not sure if I'll need this, gives all kinds of info about monitor
//static CGDirectDisplayID directDisplayId;
#else
// TODO! Windows stuff comes here
#endif

//########################################
//	local ( static ) globals
//########################################

static Boolean	gColorFlag   = TRUE;
static Boolean	gInvertFlag  = FALSE;
static Boolean	gBandFlag    = TRUE;
static Boolean	gScrollFlag  = FALSE;
static UInt8	gDirection   = 0;
static GLuint	*pnTextureID = NULL; // Array with IDs
static UInt8	gnTexID;    // current texture index
static UInt32	gnPosition; // position inside texture
static UInt8	gnLPU;      // number of lines per update

static UInt32	gDelay = SG_NORMDELAY;
static struct timeval	gLineTimeStamp = {0,0};
static struct timeval	gFrameTimeStamp = {0,0};
#ifdef SG_DEBUG
static struct timeval	gFPSTimeStamp = {0,0};
#endif


//########################################
//	exported function prototypes
//########################################

extern OSStatus iTunesPluginMainMachO( OSType inMessage, PluginMessageInfo *inMessageInfoPtr, void *refCon );

//########################################
//	static ( local ) functions
//########################################

/* Set time reference */
static inline void startuSec( struct timeval *tv )
{
	gettimeofday(tv, NULL);
}

/* Return number of microseconds since last call of startuSec.
 * This will overflow after 1 hour, 11 minutes and 34.967 seconds. */
static inline UInt32 getuSec( struct timeval tv )
{
	UInt32 result;
	struct timeval timeNow;
	gettimeofday(&timeNow, NULL);
	result = timeNow.tv_sec-tv.tv_sec;
	return 1000000*result+(timeNow.tv_usec-tv.tv_usec);
}


/* Returns the nearest speed menu setting to the current setting */
static UInt8 delayToSpeed( UInt32 delay )
{
	/* As with many things, a logarithmic scale is better here than a linear one.
	 * It's a bit stupid to recalculate the sqrts every time because they're known
	 * at compile time. But hey, this code is not executed 1000 times per second. */
	if( delay <= SG_MINDELAY )
		return 1;
	else if( delay <= sqrt(SG_FASTDELAY*SG_NORMDELAY) )
		return 2;
	else if( delay <= sqrt(SG_NORMDELAY*SG_SLOWDELAY) )
		return 3;
	else
		return 4;
}


//########################################
//	MyMemClear
//########################################
static void MyMemClear( LogicalAddress dest, SInt32 length )
{
	register unsigned char	*ptr;

	ptr = ( unsigned char* ) dest;
	
	while ( length-- > 0 )
		*ptr++ = 0;
}	// MyMemClear


#ifdef SG_DEBUG
static int checkGLError( const char *funcName )
{
 	GLenum suckage = glGetError();
	if( suckage == GL_NO_ERROR ) {
		//fprintf(stderr, "%s says GL_NO_ERROR\n", funcName ); // DEBUG
		return 0;
	}
	if( suckage == GL_INVALID_ENUM )
		fprintf(stderr, "%s barfs: GL_INVALID_ENUM\n", funcName );
	else if( suckage == GL_INVALID_VALUE )
		fprintf(stderr, "%s barfs: GL_INVALID_VALUE\n", funcName );
	else if( suckage == GL_INVALID_OPERATION )
		fprintf(stderr, "%s barfs: GL_INVALID_OPERATION\n", funcName );
	else if( suckage == GL_STACK_OVERFLOW )
		fprintf(stderr, "%s barfs: GL_STACK_OVERFLOW\n", funcName );
	else if( suckage == GL_STACK_UNDERFLOW )
		fprintf(stderr, "%s barfs: GL_STACK_UNDERFLOW\n", funcName );
	else if( suckage == GL_OUT_OF_MEMORY )
		fprintf(stderr, "%s barfs: GL_OUT_OF_MEMORY\n", funcName );
	else
		fprintf(stderr, "%s barfs an unknown error: %d\n", funcName, (int)suckage );
	return 1;
}
#endif


//########################################
// setupTextures
//########################################
static void setupTextures( void )
{
	unsigned int i;
	unsigned char *blankTexture = malloc(SG_TEXWIDTH*SG_TEXHEIGHT*4);
	memset( blankTexture, 0, SG_TEXWIDTH*SG_TEXHEIGHT*4 );
	if(pnTextureID)
		free(pnTextureID);
	// Strictly spoken, we only need to alloc enough textures to cover the viewport,
	// but this would make resizing complicated.
	pnTextureID = malloc(SG_NTEXTURES*sizeof(GLuint));
	glEnable(GL_TEXTURE_2D);
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
	glGenTextures(SG_NTEXTURES, pnTextureID);
#ifdef SG_DEBUG
	checkGLError("glGenTextures");
#endif
	for( i=0; i<SG_NTEXTURES; i++ ) {
		glBindTexture(GL_TEXTURE_2D, pnTextureID[i]);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		// Turn off texture repeating
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
		
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB,    //target, LOD (for mipmaps), internalFormat
		             SG_TEXWIDTH, SG_TEXHEIGHT,   //w, h
		             0, GL_BGRA, ARGB_IMAGE_TYPE, //border, format, type
		             blankTexture);               //data
	}
	free(blankTexture);
	glEnable(GL_TEXTURE_2D);
	gnPosition = 0;
	gnTexID = 0;
}

//########################################
// initOpenGL
//########################################
static void initOpenGL( void )
{
#if TARGET_OS_MAC
	AGLPixelFormat fmt;
	GLint attrib[] = {AGL_RGBA, AGL_RED_SIZE, 8, AGL_GREEN_SIZE, 8, AGL_BLUE_SIZE, 8, AGL_ALPHA_SIZE, 8, AGL_ACCELERATED, AGL_DOUBLEBUFFER, AGL_DEPTH_SIZE, 24, AGL_CLOSEST_POLICY, AGL_NONE};
	
	//directDisplayId = CGMainDisplayID();
	fmt = aglChoosePixelFormat(NULL, 0, attrib);
	// should test if fmt == NULL
	// create AGL context
	myContext = aglCreateContext(fmt, NULL);
	// should test if myContext == NULL
	aglDestroyPixelFormat(fmt);
#else
	// TODO: add stuff for Windows WGL
#endif
}	

//########################################
// ProcessRenderData
//########################################
static void ProcessRenderData( VisualPluginData *visualPluginDataPtr, const RenderVisualData *renderData )
{
//	SInt16		index;
//	SInt32		channel;

	if(renderData == nil) {
		MyMemClear(&visualPluginDataPtr->renderData,sizeof(visualPluginDataPtr->renderData));
		return;
	}

	visualPluginDataPtr->renderData = *renderData;
	
	/* Anti-banding: we assume there is no frequency content in the highest
	 * frequency possible (which should be the case for all normal music).
	 * So if there is something there, we subtract it from all frequencies. */
	if(gBandFlag) {
		UInt8 *spectrumDataL = visualPluginDataPtr->renderData.spectrumData[0],
		      *spectrumDataR = visualPluginDataPtr->renderData.spectrumData[1];
		SInt16 i;
		SInt16 biasL = spectrumDataL[kVisualNumSpectrumEntries/2-1],
		       biasR = spectrumDataR[kVisualNumSpectrumEntries/2-1];
		for( i=0; i<kVisualNumSpectrumEntries/2; i++ ) {
			spectrumDataL[i] -= (spectrumDataL[i]-biasL > 0) ? biasL : spectrumDataL[i];
			spectrumDataR[i] -= (spectrumDataR[i]-biasR > 0) ? biasR : spectrumDataR[i];
		}
	}		
	
	/* This just finds the min & max values of the spectrum data, if
	 * there's no need for this, you can drop this to save some CPU */
/*	for (channel = 0;channel < renderData->numSpectrumChannels;channel++)
	{
		visualPluginDataPtr->minLevel[channel] = 
			visualPluginDataPtr->maxLevel[channel] = 
			renderData->spectrumData[channel][0];

		for (index = 1; index < kVisualNumSpectrumEntries; index++)
		{
			UInt8		value;
			
			value = renderData->spectrumData[channel][index];

			if (value < visualPluginDataPtr->minLevel[channel])
				visualPluginDataPtr->minLevel[channel] = value;
			else if (value > visualPluginDataPtr->maxLevel[channel])
				visualPluginDataPtr->maxLevel[channel] = value;
		}
	}*/
}


static void UInt8ToARGB(UInt8 pValueL, UInt8 pValueR, UInt8 *pARGBPtr)
{
	if(gInvertFlag) {
		pValueL = 0xFF-pValueL;
		pValueR = 0xFF-pValueR;
	}
	
	pARGBPtr[0] = 0xFF;
	if(gColorFlag)	{
		pARGBPtr[1] = pValueL;
		pARGBPtr[2] = pValueR;
		pARGBPtr[3] = 0x00;
	}
	else
		pARGBPtr[3] = pARGBPtr[2] = pARGBPtr[1] = (pValueL+pValueR)/2;
}

static UInt8 freshPixels[SG_TEXWIDTH*4*SG_MAXCHUNK];
static int nStored = 0;
#ifdef SG_DEBUG
static unsigned int nFps = 0;
#endif

//############################################
//	RenderVisualPort, aka drawing happens here
//############################################
static void RenderVisualPort(VisualPluginData *visualPluginData, GRAPHICS_DEVICE destPort,const Rect *destRect,Boolean onlyUpdate)
{
	(void) visualPluginData;
	(void) onlyUpdate;
	int i;
	UInt8 *spectrumDataL = visualPluginData->renderData.spectrumData[0],
	      *spectrumDataR = visualPluginData->renderData.spectrumData[1];
	UInt16 nTimePixels;
	
	if (destPort == nil)
		return;

	glClearColor( 0.0, 0.0, 0.0, 0.0 );
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
	// Update the texture and only draw if chunk is full, otherwise exit
	for( i=0; i<SG_TEXWIDTH; i++ )
		UInt8ToARGB( spectrumDataL[i], spectrumDataR[i], &freshPixels[(nStored*SG_TEXWIDTH+i)*4] );
	nStored++;
	if( nStored < gnLPU )
		return;
	nStored = 0;
	
	/* glTexSubImage2D is a huge bottleneck!
	 * Updating larger chunks is slightly more efficient overall, but the incurred delay
	 * quickly becomes so large that it causes an unacceptable gap in the data.
	 * TODO: tune TEXHEIGHT for optimal speed
	 * TODO 2: check if using glCopyTexSubImage2D is more efficient, we
	 *         want to squeeze the absolute max performance out of this!!! */
	glBindTexture( GL_TEXTURE_2D, pnTextureID[gnTexID] );
	glPixelStorei(GL_UNPACK_ALIGNMENT, 4);
	glPixelStorei(GL_UNPACK_ROW_LENGTH, SG_TEXWIDTH);
	glTexSubImage2D( GL_TEXTURE_2D, 0, 0, gnPosition, SG_TEXWIDTH,
	                 gnLPU, GL_BGRA, ARGB_IMAGE_TYPE, freshPixels );	
	gnPosition = gnPosition+gnLPU;

	if(gDirection == 0)
		nTimePixels = visualPluginData->destRect.right - visualPluginData->destRect.left;
	else
		nTimePixels = visualPluginData->destRect.bottom - visualPluginData->destRect.top;
	int nNumTiles = (int)ceil((float)nTimePixels/SG_TEXHEIGHT);
	if(gScrollFlag)
		nNumTiles++;

	if( !gScrollFlag && gnTexID*SG_TEXHEIGHT+gnPosition >= nTimePixels ) {
		gnTexID = 0;
		gnPosition = 0;
	}
	else if( gnPosition >= SG_TEXHEIGHT ) {
		gnPosition = 0;
		gnTexID++;
		if( gScrollFlag && gnTexID >= nNumTiles )
			gnTexID = 0;
	}

	float fTexWid = SG_TEXHEIGHT*2.0/nTimePixels;
	if(!gScrollFlag) {
		for( i=0; i<nNumTiles; i++ ) {
			glBindTexture( GL_TEXTURE_2D, pnTextureID[i] );
			glBegin(GL_QUADS);
			if(gDirection == 0) { // Horizontal
				float fLeft = -1.0f+i*fTexWid,
				      fRight = -1.0f+(i+1)*fTexWid;
				glTexCoord2f(0.0f, 0.0f);
				glVertex3f( fLeft, -1.0f, 0.0f);
				glTexCoord2f(1.0f, 0.0f);
				glVertex3f( fLeft, 1.0f, 0.0f);
				glTexCoord2f(1.0f, 1.0f);
				glVertex3f( fRight, 1.0f, 0.0f);
				glTexCoord2f(0.0f, 1.0f);
				glVertex3f( fRight, -1.0f, 0.0f);
			}
			else {
				float fTop = 1.0f-i*fTexWid,
				      fBtm = 1.0f-(i+1)*fTexWid;
				glTexCoord2f(0.0f, 0.0f);
				glVertex3f( -1.0f, fTop, 0.0f);
				glTexCoord2f(1.0f, 0.0f);
				glVertex3f(  1.0f, fTop, 0.0f);
				glTexCoord2f(1.0f, 1.0f);
				glVertex3f(  1.0f, fBtm, 0.0f);
				glTexCoord2f(0.0f, 1.0f);
				glVertex3f( -1.0f, fBtm, 0.0f);		
			}
			glEnd();
		}
	}
	else {
		float fOffset = gnPosition*2.0/nTimePixels;
		for( i=0; i<nNumTiles; i++ ) {
			UInt8 nTexID = (gnTexID-i+nNumTiles)%nNumTiles;
			glBindTexture( GL_TEXTURE_2D, pnTextureID[nTexID] );
			glBegin(GL_QUADS);
			if(gDirection == 0) { // Horizontal
				float fLeft  = 1.0f-i*fTexWid-fOffset,
				      fRight = 1.0f+(1-i)*fTexWid-fOffset;
				glTexCoord2f(0.0f, 0.0f);
				glVertex3f( fLeft, -1.0f, 0.0f);
				glTexCoord2f(1.0f, 0.0f);
				glVertex3f( fLeft, 1.0f, 0.0f);
				glTexCoord2f(1.0f, 1.0f);
				glVertex3f( fRight, 1.0f, 0.0f);
				glTexCoord2f(0.0f, 1.0f);
				glVertex3f( fRight, -1.0f, 0.0f);
			}
			else { // Vertical
				float fTop = -1.0f+i*fTexWid+fOffset,
				      fBtm = -1.0f+(i-1)*fTexWid+fOffset;
				glTexCoord2f(0.0f, 0.0f);
				glVertex3f( -1.0f, fTop, 0.0f);
				glTexCoord2f(1.0f, 0.0f);
				glVertex3f(  1.0f, fTop, 0.0f);
				glTexCoord2f(1.0f, 1.0f);
				glVertex3f(  1.0f, fBtm, 0.0f);
				glTexCoord2f(0.0f, 1.0f);
				glVertex3f( -1.0f, fBtm, 0.0f);		
			}
			glEnd();
		}
	}
	glBindTexture( GL_TEXTURE_2D, 0 ); // unbind texture
		
	glFinish();
	glFlush();
	
	#if TARGET_OS_MAC
	aglSwapBuffers(myContext);
	#else
	{
		HDC hdc = GetDC(destPort);
		SwapBuffers(hdc);
	}
	#endif
	
	UInt32 nUSec = getuSec(gFrameTimeStamp);
	// We try to maintain a framerate of about 60FPS. If it drops below that,
	// we update fewer lines at once to produce a smoother display.
	if( gnLPU > 1 && nUSec > 16000 )
		gnLPU /= 2;
	// If the framerate is more than 90FPS, we update more lines at once,
	// which allows to achieve an even higher lines-per-second rate.
	// But, we must keep the update blocks aligned.
	else if( gnLPU < SG_MAXCHUNK && nUSec < 11000 && (gnPosition % (gnLPU*2) == 0) )
		gnLPU *= 2;
	startuSec(&gFrameTimeStamp);
	
#ifdef SG_DEBUG
	if( getuSec(gFPSTimeStamp) < 1000000 )
		nFps++;
	else {
		fprintf(stderr, "SpectroGraph fps: %u, LPU: %u, delay: %lu; NT: %d\n", nFps, gnLPU, nUSec, nNumTiles );
		startuSec(&gFPSTimeStamp);
		nFps = 0;
	}
#endif
}


// ChangeVisualPort
//
static OSStatus ChangeVisualPort(VisualPluginData *visualPluginData, GRAPHICS_DEVICE destPort, const Rect *destRect)
{
	OSStatus		status;

	status = noErr;
			
	visualPluginData->destPort = destPort;
	if(destRect != nil) {
		GLint bufferRect[4];
		GLsizei viewportWidth, viewportHeight;
		Rect portRect;
		visualPluginData->destRect = *destRect;
		
#ifdef SG_DEBUG
		if( destPort == NULL )
			fprintf(stderr, "ChangeVisualPort: destPort==NULL even though destRect!=NULL, have a nice crash!\n" );
#endif
		
		// TODO: add tests on results (GLboolean)
		aglSetDrawable(myContext, destPort);
		aglSetCurrentContext(myContext);
		// By setting AGL_SWAP_INTERVAL to 1, the swapping of buffers is synchronized with monitor retraces.
		GLint swapinterval = 0;
		aglSetInteger(myContext, AGL_SWAP_INTERVAL, &swapinterval);	
		
		// Set the bufferRect, which seems to be AGL jargon for the viewport.
		// The bufferRect must be specified as left, bottom, width, height.
		// TODO: this is way more complex in windoze, check VisualGraphicsCore::getCanvasSurroundingRect
		GetPortBounds(destPort, &portRect); // = canvasSurroundingRect (canvasRect = destRect)
		bufferRect[0] = (GLint)destRect->left;
		bufferRect[1] = (GLint)(portRect.bottom - portRect.top - destRect->bottom);
		bufferRect[2] = (GLint)(destRect->right - destRect->left);
		bufferRect[3] = (GLint)(destRect->bottom - destRect->top);
		//fprintf(stderr, "ChangeVisualPort: portBounds=%d,%d,%d,%d\n", portRect.left, portRect.top, portRect.right, portRect.bottom ); // DEBUG
		//fprintf(stderr, "ChangeVisualPort: bufferRect=%ld,%ld,%ld,%ld\n", bufferRect[0], bufferRect[1], bufferRect[2], bufferRect[3] ); // DEBUG
		aglEnable(myContext, AGL_BUFFER_RECT);
		aglSetInteger(myContext, AGL_BUFFER_RECT, bufferRect);
		
		// Notify the rendering context that the window geometry has changed
		aglUpdateContext(myContext);
		
		// FIXME: some crap needs to be added for windoze if fullscreen. Check VisualGraphicsCore::setViewport
		viewportWidth = destRect->right - destRect->left;
		viewportHeight = destRect->bottom - destRect->top;
		glViewport(0, 0, viewportWidth, viewportHeight);
		
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		
		glOrtho(-1.0, 1.0, -1.0, 1.0, 0.0, 10.0);
		
		glMatrixMode(GL_MODELVIEW);
		
		// TODO (cosmetic): if viewport is enlarged, erase new texture parts
	}
	
	return status;
}

/*
	ResetRenderData
*/
static void ResetRenderData(VisualPluginData *visualPluginData)
{
	MyMemClear(&visualPluginData->renderData,sizeof(visualPluginData->renderData));

	visualPluginData->minLevel[0] = 
		visualPluginData->minLevel[1] =
		visualPluginData->maxLevel[0] =
		visualPluginData->maxLevel[1] = 0;
}


/*
  Restart drawing (only in scan mode)
 */
void rewindDisplay()
{
	if(!gScrollFlag) {
		gnPosition = 0;
		gnTexID = 0;
	}
}


#if TARGET_OS_MAC
/* 
	settingsControlHandler
*/
pascal OSStatus settingsControlHandler(EventHandlerCallRef inRef,EventRef inEvent, void* userData)
{
	WindowRef wind=NULL;
	ControlID controlID;
	ControlRef control=NULL;
	//get control hit by event
	GetEventParameter(inEvent,kEventParamDirectObject,typeControlRef,NULL,sizeof(ControlRef),NULL,&control);
	wind=GetControlOwner(control);
	GetControlID(control,&controlID);
	switch(controlID.id) {
		case kColorSettingID:
			gColorFlag = GetControlValue(control);
			break;
		case kInvertSettingID:
			gInvertFlag = GetControlValue(control);
			break;
		case kBandSettingID:
			gBandFlag = GetControlValue(control);
			break;
		case kScrollSettingID:
			gScrollFlag = GetControlValue(control);
			break;
		case kDirSettingID:
			gDirection = GetControlValue(control)-1;
			break;
		case kSpeedSettingID:
		{
			UInt32 speeds[5] = {0, SG_MINDELAY, SG_FASTDELAY, SG_NORMDELAY, SG_SLOWDELAY};
			gDelay = speeds[GetControlValue(control)];
		}
			break;
		case kOKSettingID:
			HideWindow(wind);
			break;
	}
	return noErr;
}
#endif

/*
	VisualPluginHandler
*/
static OSStatus VisualPluginHandler(OSType message,VisualPluginMessageInfo *messageInfo,void *refCon)
{
	OSStatus			status;
	VisualPluginData *	visualPluginData;

	visualPluginData = (VisualPluginData*) refCon;
	
	status = noErr;

	switch (message)
	{
		/*
		 * Apple says:
			Sent when the visual plugin is registered.  The plugin should do minimal
			memory allocations here.  The resource fork of the plugin is still available.
		 * I say:
		    iTunes will 'register' each plugin when iTunes is started. Even though your plug-in
			will not necessarily be shown on this occasion, you can still do some initializations.
			However, don't do anything that will hog memory or take ages.
		*/		
		case kVisualPluginInitMessage:
		{
			visualPluginData = (VisualPluginData*) calloc(1, sizeof(VisualPluginData));
			if (visualPluginData == nil)
			{
				status = memFullErr;
				break;
			}

			visualPluginData->appCookie	= messageInfo->u.initMessage.appCookie;
			visualPluginData->appProc	= messageInfo->u.initMessage.appProc;

			messageInfo->u.initMessage.refCon	= (void*) visualPluginData;
#ifdef SG_DEBUG
			fprintf(stderr, "SpectroGraph inited\n");
#endif
			break;
		}
			
		/*
			Sent when the visual plugin is unloaded
		*/		
		case kVisualPluginCleanupMessage:
#ifdef SG_DEBUG
			fprintf(stderr, "Unloading SpectroGraph...\n");
#endif
			if (visualPluginData != nil)
				free(visualPluginData);
			break;
			
		/*
			Sent when the visual plugin is enabled.  iTunes currently enables all
			loaded visual plugins.  The plugin should not do anything here.
		*/
		case kVisualPluginEnableMessage:
		case kVisualPluginDisableMessage:
			break;

		/*
			Sent if the plugin requests idle messages.  Do this by setting the kVisualWantsIdleMessages
			option in the RegisterVisualMessage.options field.
		*/
		case kVisualPluginIdleMessage:
			/* This is where it gets nasty. Idle messages can be sent at any time: while iTunes is playing
			 * (frequently), when paused (constantly), and even when the visualizer is off (a few times
			 * per second). Moreover, _all_ plug-ins receive idle messages even if another one is active.
			 * Because I used 0xFFFFFFFF for timeBetweenData, iTunes will use 100% cpu, both during playback
			 * and while paused. This is why I included the usleep calls. Mind that usleep will pause the
			 * _entire_ iTunes process, so a sensible value must be used (1msec doesn't seem to interfere
			 * with normal operation).
			 * I need to check if iTunes is paused to avoid messing up the timing of the rendering routine,
			 * and check if the plug-in is active to avoid rendering a non-existing port. */
			if( visualPluginData->playing == false && visualPluginData->destPort != nil ) {
				if( getuSec(gLineTimeStamp) > gDelay ) {
					startuSec(&gLineTimeStamp);
					RenderVisualPort(visualPluginData,visualPluginData->destPort,&visualPluginData->destRect,false);
				}
				else
					usleep(SG_USLEEP); // TODO: find Windows equivalent
			}
			break;
			
		/*
			Sent if the plugin requests the ability for the user to configure it.  Do this by setting
			the kVisualWantsConfigure option in the RegisterVisualMessage.options field.
		*/
#if TARGET_OS_MAC					
		case kVisualPluginConfigureMessage:
		{
			static EventTypeSpec controlEvent={kEventClassControl,kEventControlHit};
			static const ControlID kColorSettingControlID ={'cbox',kColorSettingID};
			static const ControlID kInvertSettingControlID={'cbox',kInvertSettingID};
			static const ControlID kBandSettingControlID  ={'cbox',kBandSettingID};
			static const ControlID kScrollSettingControlID={'cbox',kScrollSettingID};
			static const ControlID kDirSettingControlID   ={'popm',kDirSettingID};
			static const ControlID kSpeedSettingControlID ={'popm',kSpeedSettingID};
			
			static WindowRef settingsDialog;
			static ControlRef color =NULL;
			static ControlRef invert=NULL;
			static ControlRef band  =NULL;
			static ControlRef scroll=NULL;
			static ControlRef dirm  =NULL;
			static ControlRef speedm=NULL;
			
			IBNibRef 		nibRef;
			//we have to find our bundle to load the nib inside of it
			CFBundleRef iTunesPlugin;
			
			iTunesPlugin=CFBundleGetBundleWithIdentifier(CFSTR("be.dr-lex.SpectroGraph"));
			if( iTunesPlugin == NULL ) {
				fprintf( stderr, "SpectroGraph error: could not find bundle\n" );
				SysBeep(2);
			}
			else {
				CreateNibReferenceWithCFBundle(iTunesPlugin,CFSTR("SettingsDialog"), &nibRef);
			
				if( nibRef != nil ) {
					CreateWindowFromNib(nibRef, CFSTR("PluginSettings"), &settingsDialog);
					DisposeNibReference(nibRef);
				
					if(settingsDialog) {
						InstallWindowEventHandler(settingsDialog,NewEventHandlerUPP(settingsControlHandler),
						                          1,&controlEvent,0,NULL);
						GetControlByID(settingsDialog,&kColorSettingControlID, &color);
						GetControlByID(settingsDialog,&kInvertSettingControlID,&invert);
						GetControlByID(settingsDialog,&kBandSettingControlID,  &band);
						GetControlByID(settingsDialog,&kScrollSettingControlID,&scroll);
						GetControlByID(settingsDialog,&kDirSettingControlID,   &dirm);
						GetControlByID(settingsDialog,&kSpeedSettingControlID, &speedm);
						
						SetControlValue(color, gColorFlag);
						SetControlValue(invert,gInvertFlag);
						SetControlValue(band,  gBandFlag);
						SetControlValue(scroll,gScrollFlag);
						SetControlValue(dirm,  gDirection+1);
						SetControlValue(speedm,delayToSpeed(gDelay));
						ShowWindow(settingsDialog);
					}
				}
			}
		}
			break;
#endif // TARGET_OS_MAC
				
		/*
		 * Apple says:
			Sent when iTunes is going to show the visual plugin in a port.  At
			this point, the plugin should allocate any large buffers it needs.
		 * I say:
		    This message is called when the plugin is 'activated', i.e. when
			the iTunes visualizer is enabled and this plugin is selected. This
			is where you need to do all initializations you didn't do before.
		*/
		case kVisualPluginShowWindowMessage:
		{
			initOpenGL();
			
			visualPluginData->destOptions = messageInfo->u.showWindowMessage.options;
			
			status = ChangeVisualPort(	visualPluginData,
										#if TARGET_OS_WIN32
											messageInfo->u.setWindowMessage.window,
										#endif
										#if TARGET_OS_MAC
											messageInfo->u.setWindowMessage.port,
										#endif
										&messageInfo->u.showWindowMessage.drawRect);
			
			/* this HAS to be done after setting up the viewport. Otherwise it will do
			 * just nothing, it won't even produce any errors, your textures will just be white. */
			setupTextures();
			
#ifdef SG_DEBUG
			startuSec(&gFPSTimeStamp);
#endif
			if(status == noErr)
				RenderVisualPort(visualPluginData,visualPluginData->destPort,&visualPluginData->destRect,true);
#ifdef SG_DEBUG
			fprintf(stderr, "SpectroGraph started\n");
#endif
			break;
		}
		/*
		 * Apple says:
			Sent when iTunes is no longer displayed.
		 * I say:
		    Sent when the _visualizer_ is no longer displayed.  In other words:
			when the user disables the visualizer, swtiches to another visualizer,
			closes the iTunes window, or minimizes iTunes to the Dock.
		*/
		case kVisualPluginHideWindowMessage:
#ifdef SG_DEBUG
			fprintf(stderr, "Hiding SpectroGraph\n");
#endif
			(void) ChangeVisualPort(visualPluginData,nil,nil);
			
			aglSetCurrentContext(NULL);
			if( myContext != NULL ) {
				aglSetDrawable(myContext, NULL);
				aglDestroyContext(myContext);
				myContext = NULL;
			}
			
			MyMemClear(&visualPluginData->trackInfo,sizeof(visualPluginData->trackInfo));
			MyMemClear(&visualPluginData->streamInfo,sizeof(visualPluginData->streamInfo));
			break;
		
		/*
			Sent when iTunes needs to change the port or rectangle of the currently
			displayed visual.
		*/
		case kVisualPluginSetWindowMessage:
			visualPluginData->destOptions = messageInfo->u.setWindowMessage.options;

			status = ChangeVisualPort(	visualPluginData,
										#if TARGET_OS_WIN32
											messageInfo->u.showWindowMessage.window,
										#endif
										#if TARGET_OS_MAC
											messageInfo->u.showWindowMessage.port,
										#endif
										&messageInfo->u.setWindowMessage.drawRect);

			if (status == noErr)
				RenderVisualPort(visualPluginData,visualPluginData->destPort,&visualPluginData->destRect,true);
			break;
		
		/*
			Sent for the visual plugin to render a frame.
		*/
		case kVisualPluginRenderMessage:
			{
			SInt32 timeLeft = gDelay-getuSec(gLineTimeStamp); /* Overflow hazard! Hence the third test below. */
			if( (gDelay == SG_MINDELAY) || timeLeft < 0 || (UInt32)timeLeft > gDelay ) {
				startuSec(&gLineTimeStamp);
				visualPluginData->renderTimeStampID	= messageInfo->u.renderMessage.timeStampID;
				
				ProcessRenderData(visualPluginData,messageInfo->u.renderMessage.renderData);
				
				RenderVisualPort(visualPluginData,visualPluginData->destPort,&visualPluginData->destRect,false);
			}
			else if( timeLeft > SG_USLEEP*1.3 )
				usleep(SG_USLEEP);
			}
			break;
#if 0			
		/*
			Sent for the visual plugin to render directly into a port.  Not necessary for normal
			visual plugins.
		*/
		case kVisualPluginRenderToPortMessage:
			status = unimpErr;
			break;
#endif 0
		/*
			Sent in response to an update event.  The visual plugin should update
			into its remembered port.  This will only be sent if the plugin has been
			previously given a ShowWindow message.
		*/	
		case kVisualPluginUpdateMessage:
			RenderVisualPort(visualPluginData,visualPluginData->destPort,&visualPluginData->destRect,true);
			break;
		
		/*
			Sent when the player starts.
		*/
		case kVisualPluginPlayMessage:
			if (messageInfo->u.playMessage.trackInfo != nil)
				visualPluginData->trackInfo = *messageInfo->u.playMessage.trackInfoUnicode;
			else
				MyMemClear(&visualPluginData->trackInfo,sizeof(visualPluginData->trackInfo));

			if (messageInfo->u.playMessage.streamInfo != nil)
				visualPluginData->streamInfo = *messageInfo->u.playMessage.streamInfoUnicode;
			else
				MyMemClear(&visualPluginData->streamInfo,sizeof(visualPluginData->streamInfo));
		
			visualPluginData->playing = true;
			break;

		/*
			Sent when the player changes the current track information.  This
			is used when the information about a track changes,or when the CD
			moves onto the next track.  The visual plugin should update any displayed
			information about the currently playing song.
		*/
		case kVisualPluginChangeTrackMessage:
			if (messageInfo->u.changeTrackMessage.trackInfo != nil)
				visualPluginData->trackInfo = *messageInfo->u.changeTrackMessage.trackInfoUnicode;
			else
				MyMemClear(&visualPluginData->trackInfo,sizeof(visualPluginData->trackInfo));

			if (messageInfo->u.changeTrackMessage.streamInfo != nil)
				visualPluginData->streamInfo = *messageInfo->u.changeTrackMessage.streamInfoUnicode;
			else
				MyMemClear(&visualPluginData->streamInfo,sizeof(visualPluginData->streamInfo));
			break;

		/*
			Sent when the player stops.
		*/
		case kVisualPluginStopMessage:
			visualPluginData->playing = false;
			
			ResetRenderData(visualPluginData);

			RenderVisualPort(visualPluginData,visualPluginData->destPort,&visualPluginData->destRect,true);
			break;
		
		/*
			Sent when the player changes position.
		*/
		case kVisualPluginSetPositionMessage:
			break;

		/*
			Sent when the player pauses.  iTunes does not currently use pause or unpause.
			A pause in iTunes is handled by stopping and remembering the position.
		*/
		case kVisualPluginPauseMessage:
			visualPluginData->playing = false;

			ResetRenderData(visualPluginData);

			RenderVisualPort(visualPluginData,visualPluginData->destPort,&visualPluginData->destRect,true);
			break;
			
		/*
			Sent when the player unpauses.  iTunes does not currently use pause or unpause.
			A pause in iTunes is handled by stopping and remembering the position.
		*/
		case kVisualPluginUnpauseMessage:
			visualPluginData->playing = true;
			break;
		
		/*
			Sent to the plugin in response to a MacOS event.  The plugin should return noErr
			for any event it handles completely,or an error (unimpErr) if iTunes should handle it.
		*/
		#if TARGET_OS_MAC
		// TODO: what's the equivalent under Windows? Just disabling all controls is unacceptable!
		case kVisualPluginEventMessage:
			{
				EventRecord* tEventPtr = messageInfo->u.eventMessage.event;
				if ((tEventPtr->what == keyDown) || (tEventPtr->what == autoKey))
				{    // charCodeMask,keyCodeMask;
					char theChar = tEventPtr->message & charCodeMask;

					switch (theChar) {
						case	'b':
						case	'B':
							gBandFlag = !gBandFlag;
							status = noErr;
							break;
						case	'c':
						case	'C':
							gColorFlag = !gColorFlag;
							status = noErr;
							break;
						case	'h':
						case	'H':
							if(++gDirection > 1)
								gDirection = 0;
							rewindDisplay();
							status = noErr;
							break;							
						case	'i':
						case	'I':
							gInvertFlag = !gInvertFlag;
							status = noErr;
							break;
						case	'l':
						case	'L':
							gScrollFlag = !gScrollFlag;
							status = noErr;
							break;							
						case	'r':
						case	'R':
							rewindDisplay();
							status = noErr;
							break;
						case	'-':
							if( gDelay < SG_MAXDELAY )
								gDelay *= SG_FACTOR;
							else
								gDelay = SG_MAXDELAY;
							status = noErr;
							break;
						case	'+':
						case	'=':
							if( gDelay > SG_MINDELAY )
								gDelay /= SG_FACTOR;
							else
								gDelay = SG_MINDELAY;
							status = noErr;
							break;
						case	'u':
						case	'U':
							gDelay = SG_MINDELAY;
							status = noErr;
							break;
						case	'f':
						case	'F':
							gDelay = SG_FASTDELAY;
							status = noErr;
							break;
						case	'n':
						case	'N':
							gDelay = SG_NORMDELAY;
							status = noErr;
							break;
						case	's':
						case	'S':
							gDelay = SG_SLOWDELAY;
							status = noErr;
							break;
							
						default:
							status = unimpErr;
							break;
					}
				}
				else
					status = unimpErr;
			}
			break;
		#endif // TARGET_OS_MAC
		
		default:
			status = unimpErr;
			break;
	}
	return status;
}

/*
	RegisterVisualPlugin
*/
static OSStatus RegisterVisualPlugin(PluginMessageInfo *messageInfo)
{
	OSStatus			status;
	PlayerMessageInfo	playerMessageInfo;
	Str255				pluginName = kTVisualPluginName;
		
	MyMemClear(&playerMessageInfo.u.registerVisualPluginMessage,sizeof(playerMessageInfo.u.registerVisualPluginMessage));
	
	memcpy(&playerMessageInfo.u.registerVisualPluginMessage.name[0], &pluginName[0], pluginName[0] + 1);

	#if TARGET_OS_MAC					
		CFStringRef tCFStringRef = CFStringCreateWithPascalString( kCFAllocatorDefault, pluginName, kCFStringEncodingUTF8 );
		if ( tCFStringRef ) 
		{
			CFIndex length = CFStringGetLength( tCFStringRef );
			if ( length > 255 ) 
			{
				length = 255;
			}
			playerMessageInfo.u.registerVisualPluginMessage.unicodeName[0] = CFStringGetBytes( tCFStringRef, CFRangeMake( 0, length ), kCFStringEncodingUnicode, 0, FALSE, (UInt8 *) &playerMessageInfo.u.registerVisualPluginMessage.unicodeName[1], 255, NULL );
			CFRelease( tCFStringRef );
		}
	#endif TARGET_OS_MAC					

	SetNumVersion(&playerMessageInfo.u.registerVisualPluginMessage.pluginVersion,kTVisualPluginMajorVersion,kTVisualPluginMinorVersion,kTVisualPluginReleaseStage,kTVisualPluginNonFinalRelease);

	playerMessageInfo.u.registerVisualPluginMessage.options					=	kVisualWantsIdleMessages 
																				#if TARGET_OS_MAC					
																					| kVisualWantsConfigure | kVisualProvidesUnicodeName
																				#endif
																				;
	playerMessageInfo.u.registerVisualPluginMessage.handler					= (VisualPluginProcPtr)VisualPluginHandler;
	playerMessageInfo.u.registerVisualPluginMessage.registerRefCon			= 0;
	playerMessageInfo.u.registerVisualPluginMessage.creator					= kTVisualPluginCreator;
	
	/* This determines how often the plugin receives data. The name is deceiving because we can't
	 * get millisecond accuracy. Instead, ticks of 16msec are used, so it's impossible to go faster
	 * than 62.5 packets/second. For this plug-in that's still too slow, so I have to disable this
	 * by using 0xFFFFFFFF (= as fast as possible) and do my own speed control.
	 * If your own plug-in can do with 62.5 frames/second, by all means enable this, because otherwise
	 * you'll have to use similar ugly tricks as I to avoid 100% cpu usage. */	
	playerMessageInfo.u.registerVisualPluginMessage.timeBetweenDataInMS		= 0xFFFFFFFF;
	playerMessageInfo.u.registerVisualPluginMessage.numWaveformChannels		= 2;
	playerMessageInfo.u.registerVisualPluginMessage.numSpectrumChannels		= 2;
	
	playerMessageInfo.u.registerVisualPluginMessage.minWidth				= 64;
	playerMessageInfo.u.registerVisualPluginMessage.minHeight				= 64;
	playerMessageInfo.u.registerVisualPluginMessage.maxWidth				= 32767;
	playerMessageInfo.u.registerVisualPluginMessage.maxHeight				= 32767;
	playerMessageInfo.u.registerVisualPluginMessage.minFullScreenBitDepth	= 0;
	playerMessageInfo.u.registerVisualPluginMessage.maxFullScreenBitDepth	= 0;
	playerMessageInfo.u.registerVisualPluginMessage.windowAlignmentInBytes	= 0;
	
	status = PlayerRegisterVisualPlugin(messageInfo->u.initMessage.appCookie,messageInfo->u.initMessage.appProc,&playerMessageInfo);
	startuSec(&gLineTimeStamp);
	startuSec(&gFrameTimeStamp);
	gnLPU = SG_MAXCHUNK;
	
	return status;
	
}

//########################################
//	main entrypoint
//########################################

#if TARGET_OS_WIN32
IMPEXP OSStatus MAIN(OSType message,PluginMessageInfo *messageInfo,void *refCon)
#else
OSStatus iTunesPluginMainMachO(OSType message,PluginMessageInfo *messageInfo,void *refCon)
#endif
{
	OSStatus		status;
	
	(void) refCon;
	
	switch (message)
	{
		case kPluginInitMessage:
			status = RegisterVisualPlugin(messageInfo);
			break;
			
		case kPluginCleanupMessage:
			status = noErr;
			break;
			
		default:
			status = unimpErr;
			break;
	}
	
	return status;
}
